/*
   FILE:
						       	
	test_ptrace_api4_module.c
								       	
   DESCRIPTION:							       	

        Test 4 for POSIX Tracing API in RT-Linux/Linux.

	This test checks the offline trace and retrieval of events.

	In particular, the init_module creates a stream with log,
	creates a new (user) event type and then traces several events
	of this type.  Then, it creates a task which performs a flush
	operation on the stream.

	The second part of the test is a Linux program called
	"test_ptrace_api4_analyzer.c" which opens the log and
	retrieves the information from it, checking that the log is
	correct and contains all the expected information.  Please
	refer to this file for more info.

			       				
   Copyright (C) 2004
	Ana Garcia Fornes
	Andres Terrasa Barrena
	Agustin Espinosa Minguet

  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software 
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.

  LAST REVISION:
        September 2004

*/

#include <rtl.h>
#include <time.h>
#include <rtl_fifo.h>
#include <trace.h>
#include <unistd.h>

#include "test_ptrace_api4.h"

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Andres Terrasa");
MODULE_DESCRIPTION("API Test 4 for POSIX Tracing");

static int delay = 2;
static int fifo = 10;

MODULE_PARM(delay,"i");
MODULE_PARM(fifo,"i");



static int            test_failed = 0;
static trace_id_t     trid;


static int              fifo_desc;
static pthread_t        flush_task, shutdown_task;



void *task_to_flush (void *dummy) {
  int      error;

  if ((error = posix_trace_flush(trid)) != 0) {
    rtl_printf("%s: [BROK]:posix_trace_flush returned %d\n", 
	       TEST_NAME, error);
    test_failed = 1;
    return (void *) -1;
  }
  return (void *) 0;
}




void *task_to_shutdown(void *dummy) {
  struct timespec wakeup_time;
  struct timespec retry_time;
  struct posix_trace_status_info status;

  clock_gettime(CLOCK_MONOTONIC, &wakeup_time);
  wakeup_time.tv_sec += delay;
  clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME, &wakeup_time, NULL);

  posix_trace_get_status(trid, &status);
  retry_time = wakeup_time;
  while (status.posix_stream_flush_status == POSIX_TRACE_FLUSHING) {
    retry_time.tv_nsec += 100000000;
    clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME, &retry_time, NULL);
    posix_trace_get_status(trid,&status);
  }

  // OK. Flush (if any) is now over. Shut the stream down:
  posix_trace_shutdown(trid);

  return (void *) 0;
}




int init_module(void) {

  trace_attr_t   attr;
  int            error;
  int            i;
  trace_event_id_t event_string_id;
  trace_event_set_t  set1;
  char           fifo_name[16];
  event_data_struct event_data;

  // Open the fifo
  sprintf(fifo_name, "/dev/rtf%d", fifo);
  rtf_create(fifo, 100000);
  fifo_desc = open(fifo_name, O_NONBLOCK);
  if (fifo_desc < 0) {
    rtl_printf("%s: [BROK]:Cannot open %s for writing\n", 
	       TEST_NAME, fifo_name);
    return -1;
  }

  // Create the trace attribute:
  posix_trace_attr_init(&attr);

  // Set the attributes with values defined in common.h:
  posix_trace_attr_setstreamfullpolicy(&attr, MY_STREAM_POLICY);
  posix_trace_attr_setname(&attr, MY_STREAM_NAME);
  posix_trace_attr_setmaxdatasize(&attr, MY_EVENT_DATA_SIZE);
  posix_trace_attr_setstreamsize(&attr, MY_STREAM_SIZE);

  // Create the stream (with log)
  if ((error = posix_trace_create_withlog(0, &attr, fifo_desc, &trid)) != 0) {
    rtl_printf("%s: [BROK]:posix_trace_create_withlog returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }

  // Create a new user event type:
  if ((error = posix_trace_eventid_open (MY_USER_EVENT_NAME, 
					 &event_string_id)) != 0) {
    rtl_printf("%s: [BROK]:posix_trace_eventid_open returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }

  // Filter all *system* event types:
  posix_trace_eventset_fill(&set1, POSIX_TRACE_SYSTEM_EVENTS);

  if ((error = posix_trace_set_filter(trid, &set1, 
				      POSIX_TRACE_SET_EVENTSET)) != 0) { 
    rtl_printf("%s: [BROK]:posix_trace_get_filter returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }


  // Start the tracing of events: 
  if ((error = posix_trace_start(trid)) != 0) {
    rtl_printf("%s: [BROK]:posix_trace_start returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }

  // Trace an exact number of events:
  strcpy(event_data.str, MY_STRING);
  for (i = 0; i < NUMBER_OF_TRACED_EVENTS; i++) {
    event_data.serial = i;
    posix_trace_event(event_string_id, &event_data, sizeof(event_data));
  }

  // Stop the tracing of events: 
  if ((error = posix_trace_stop(trid)) != 0) {
    rtl_printf("%s: [BROK]:posix_trace_stop returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }

  // Create the task which will flush the events:
  pthread_create (&flush_task, NULL, task_to_flush, 
		  (void *) delay);

  // And the task which will shut down the stream:
  pthread_create (&shutdown_task, NULL, task_to_shutdown, NULL); 


  if (! test_failed) {
    rtl_printf("%s: PASS :Log generated. Offline trace API works OK.\n", 
	       TEST_NAME);  
  }

  return test_failed;

}





void cleanup_module(void) {
  pthread_cancel(flush_task);
  pthread_cancel(shutdown_task);

  pthread_join(flush_task, NULL);
  pthread_join(shutdown_task, NULL);
  
  // Close the fifo 
  close(fifo_desc);
  rtf_destroy(fifo);
}
