/*
   FILE:							       	
	test_ptrace_perf1_module.c							       
								       	
   DESCRIPTION:							       	

        Performance Test 1 for POSIX Tracing in RT-Linux.

	This test measures of the costs of tracing and retrieving
	events using the POSIX Tracing API.

	Costs are calculated by using the function rdtsc, which 
	measures the time in CPU clock ticks. Measurements are
	stored in a mbuff shared memory, from where a Linux program
	retrieves them and store them in some files. These files
	are later analyzed in order to compute some basic
	statistics (minimum, maximum, average and variance values).

	This module has three entry parameters:

	- period : period of the tasks tracing and retrieving the
	           events for measurement.

	- data_size: Size in bytes of the data being traced and 
	           retrieved along with each event. Valid values
		   are between 0 and 4096 bytes.

        - policy: full policy to be used by the stream.

	This makes easier to apply the performance test in different
	scenarios.

					       				
   Copyright (C) 2003,2004
	Andres Terrasa Barrena
	Ana Garcia Fornes
	Agustin Espinosa Minguet

  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software 
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.

  LAST REVISION:
        October, 2004

	(This test is based on a previous one made in Dec. 2003)
*/


#include <rtl.h>
#include <time.h>
#include <pthread.h>
#include <rtl_sched.h>
#include <trace.h>

#include "test_ptrace_perf1.h"


MODULE_LICENSE("GPL");
MODULE_AUTHOR("Andres Terrasa");
MODULE_DESCRIPTION("Performance Test 1 for POSIX Tracing");

static long period    = 1000000;
static int  data_size = 4;
static int  policy    = 1; // 1 => LOOP, 2 => UNTIL_FULL

MODULE_PARM(period,"l");
MODULE_PARM(data_size,"i");
MODULE_PARM(policy,"i");


#define TEST_NAME "ptrace_test_perf1"


#define MAX_DATASIZE  4096

static char event_data[MAX_DATASIZE];

static trace_id_t       trid;
static trace_event_id_t user_event;
static pthread_t        thr_reader, thr_writer;
static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;

static int mu_in = 0; // Index in the time array (mbuff)




void *writer(void *writer_period) {
  int   i;
  int   k;
  //  void  *data;
  long            task_period;
  struct timespec next_period;
  long long before;
  long long after1;
  long long after2;

  task_period = (long) * ((long *) writer_period);


  clock_gettime(CLOCK_MONOTONIC, &next_period);
  timespec_add_ns(&next_period, task_period);

  k = 0;

  for (i=0; i<20000; i++) {

    clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME, &next_period, NULL);
    timespec_add_ns(&next_period, task_period);

    pthread_mutex_lock(&mutex);

    rdtscll(before);
    posix_trace_event(user_event, event_data, data_size);
    rdtscll(after1);
    posix_trace_event(user_event, event_data, data_size);
    rdtscll(after2);

    if (mu_in < MAX_MEASUREMENTS) {
      measure_shm[mu_in].type = TRACE_EVENT_1;
      measure_shm[mu_in].before = before;
      measure_shm[mu_in].after = after1;
      mu_in++;
    }
    if (mu_in < MAX_MEASUREMENTS) {
      measure_shm[mu_in].type = TRACE_EVENT_2;
      measure_shm[mu_in].before = after1;
      measure_shm[mu_in].after = after2;
      mu_in++;
    }

    pthread_mutex_unlock(&mutex);
    
  }
  return (void *) 0;
}








void *reader(void *reader_period) {
  int             error;
  struct posix_trace_event_info event;
  char            data[MAX_DATASIZE];
  size_t          datalen;
  int             unavailable1;
  int             unavailable2;
  long            task_period;
  struct timespec next_period;
  long long       before;
  long long       after1;
  long long       after2;

  task_period = (long) * ((long *) reader_period);

  clock_gettime(CLOCK_MONOTONIC, &next_period);
  timespec_add_ns(&next_period, task_period);
  timespec_add_ns(&next_period, (long) task_period/2);

  error = 0; unavailable1 = unavailable2 = 0;

  while (! error) {
    
    clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME, &next_period, NULL);
    timespec_add_ns(&next_period, period);

    pthread_mutex_lock(&mutex);

    rdtscll(before);
    error = posix_trace_trygetnext_event(trid, &event, &data, data_size, 
					 &datalen, &unavailable1);
    
    rdtscll(after1);
    error = posix_trace_trygetnext_event(trid, &event, &data, data_size, 
					 &datalen, &unavailable2);
    
    rdtscll(after2);
    
    if (mu_in < MAX_MEASUREMENTS && ! unavailable1) {
      measure_shm[mu_in].type = GETNEXT_EVENT_1;
      measure_shm[mu_in].before = before;
      measure_shm[mu_in].after = after1;
      mu_in++;
    }
    if (mu_in < MAX_MEASUREMENTS && ! unavailable2) {
      measure_shm[mu_in].type = GETNEXT_EVENT_2;
      measure_shm[mu_in].before = after1;
      measure_shm[mu_in].after = after2;
      mu_in++;
    }

    pthread_mutex_unlock(&mutex);

  }

  return (void *) 0;
}




 


int init_module(void) {

  trace_attr_t      attr;
  pthread_attr_t    thattr;
  trace_event_set_t set;
  int               error;
  size_t            struct_size;
  int               full_policy;


  // Create the shared memory area to communcate with the Linux app.
  struct_size = sizeof(measure_unit_t);
  measure_shm = (volatile measure_unit_t *) 
    mbuff_alloc(MEASUREMENTS_MBUF, struct_size * MAX_MEASUREMENTS);

  if(measure_shm == NULL) {
    rtl_printf("Error: mbuff_alloc failed\n");
    return -1;	
  }

  // Check if data_size is greater than the max data chunk we're tracing:
  if (data_size < 0 || data_size > MAX_DATASIZE) {
    rtl_printf("Error: entry parameter data_size (%d) greater than %d\n", 
	       data_size, MAX_DATASIZE);
    return -1;
  }

  // Create and set the trace attribute:
  error = posix_trace_attr_init(&attr);
  
  full_policy = (policy == 1 ? POSIX_TRACE_LOOP : POSIX_TRACE_UNTIL_FULL);
  error = posix_trace_attr_setstreamfullpolicy (&attr, full_policy);

 // Create the stream:
  error = posix_trace_create(0, &attr, &trid); 
  if (error) {
    rtl_printf("Error: Cannot create the stream\n");
    return -1;
  }

 
  // Create new event types associated with this stream:
  error = posix_trace_eventid_open ("user event", &user_event);


  // Set the stream filter to only record user events:
  posix_trace_eventset_fill(&set, POSIX_TRACE_SYSTEM_EVENTS);

  posix_trace_set_filter(trid, (const trace_event_set_t *) &set, 
			 POSIX_TRACE_SET_EVENTSET);

  // Start tracing:
  posix_trace_start(trid);

  //  rtl_printf("Creating tasks with period = %ld\n", period);
	
  // Create the tasks: 
  pthread_attr_init (&thattr); 
  pthread_create (&thr_writer, &thattr, writer, (void *) &period); 
  pthread_create (&thr_reader, &thattr, reader, (void *) &period); 

  /* No errors. All tests were fine. Return 0. */
  rtl_printf("%s: PASS :Module completed. All measurements collected.\n", 
	     TEST_NAME);  
  return 0;
}



void cleanup_module(void) {

  // Stop and shutdown the stream:
  posix_trace_shutdown(trid); 

  // Cancel the tasks:
  pthread_cancel(thr_reader); 
  pthread_cancel(thr_writer); 
  pthread_join(thr_reader,NULL); 
  pthread_join(thr_writer,NULL); 

  // Free the shared memory area
  mbuff_free(MEASUREMENTS_MBUF,(void*)measure_shm);
}
