/*
   FILE:
						       	
	test_ptrace_stress2_module.c
								       	
   DESCRIPTION:							       	

        Stress test No. 2 POSIX Tracing in RT-Linux/Linux.

	This test puts the Tracing support under maximum performance
	requirements: the tracing of events is done in a single, un-
	interrupted burst of MAX_EVENTS events, and all these events
	are traced into the maximum number of streams simultaneously.
	Half of the available streams are with log, half without log.

	If the POSIX Tracing support has not been configured to use
	less than its maximum, this means tracing to 8 simultaneous
	streams.

	The test is successfull if:
	
	  a) The RT-Linux kernel does not hang :)
	  b) All pthread reading the streams w/o log retrieve all
	     the events correctly
	  c) The off-line Linux analyzers open and retrieve all the
	     events stored in streams with log, from their corresponding
	     logs.

			       				
   Copyright (C) 2004
	Ana Garcia Fornes
	Andres Terrasa Barrena
	Agustin Espinosa Minguet

  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software 
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.

  LAST REVISION:
        October 2004

*/

#include <rtl.h>
#include <time.h>
#include <rtl_fifo.h>
#include <trace.h>
#include <unistd.h>

#include "test_ptrace_stress2.h"

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Andres Terrasa");
MODULE_DESCRIPTION("Stress test 2 for POSIX Tracing");



static int delay = 2;
static int fifo1 = 10;
static int fifo2 = 11;
static int fifo3 = 12;
static int fifo4 = 13;

MODULE_PARM(delay,"i");
MODULE_PARM(fifo1,"i");
MODULE_PARM(fifo2,"i");
MODULE_PARM(fifo3,"i");
MODULE_PARM(fifo4,"i");


static int            test_failed = 0;
static trace_id_t     trid1, trid2, trid3, trid4;
static trace_id_t     trid5, trid6, trid7, trid8;


static int              fifo1_desc, fifo2_desc;
static int              fifo3_desc, fifo4_desc;
static pthread_t        shutdown_task1, shutdown_task2;
static pthread_t        shutdown_task3, shutdown_task4;
static pthread_t        reader_task1, reader_task2;
static pthread_t        reader_task3, reader_task4;




void *task_reader(void *trid_number) {
  trace_id_t      trid;
  int             nevents;
  int             error;
  struct posix_trace_event_info event;
  char            event_name[TRACE_NAME_MAX];
  size_t          datalen;
  int             unavailable;
  event_data_struct event_data;

  trid = *( (trace_id_t *) trid_number);

  error = 0; unavailable = 0;
  nevents = 0;

  while (! error && ! unavailable) {

    error = posix_trace_trygetnext_event(trid, &event, 
					 &event_data, sizeof(event_data), 
					 &datalen, &unavailable);

    if (error || unavailable) {
      break;
    }

    posix_trace_eventid_get_name (trid, event.posix_event_id, event_name);
    if (strcmp(event_name,"user event string")) {
      rtl_printf("%s: [BROK]:Retrieved event of wrong type\n", 
		 TEST_NAME);
      test_failed = 1;
      return (void *) -1;
    }
  
    if (strcmp((char *) event_data.str, MY_STRING)) {
      rtl_printf("%s: [BROK]:Retrieved event with wrong data\n", 
		 TEST_NAME);
      test_failed = 1;
      return (void *) -1;
    }

    if (event_data.serial != nevents) {
      rtl_printf("%s: [BROK]:Event in stream has wrong serial data\n",
	      TEST_NAME);
      test_failed = 1;
      return (void *) -1;
    }
    nevents += 1;


  } /* end while */

  if (nevents < NUMBER_OF_TRACED_EVENTS) {
    rtl_printf("%s: [BROK]:Retrieved %d events when %d expected\n", 
	       TEST_NAME, nevents, NUMBER_OF_TRACED_EVENTS);
     test_failed = 1;
    return (void *) -1;
  }

  /*  rtl_printf("%s: [INFO] Task successfully retrieved all events. Done!\n", 
      TEST_NAME);*/
  return (void *) 0;
}








void *task_to_shutdown(void *trid_number) {
  struct timespec wakeup_time;
  struct timespec retry_time;
  struct posix_trace_status_info status;
  trace_id_t trid;

  trid = *( (trace_id_t *) trid_number);

  clock_gettime(CLOCK_MONOTONIC, &wakeup_time);
  wakeup_time.tv_sec += delay;
  clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME, &wakeup_time, NULL);

  posix_trace_get_status(trid, &status);
  retry_time = wakeup_time;
  while (status.posix_stream_flush_status == POSIX_TRACE_FLUSHING) {
    retry_time.tv_nsec += 100000000;
    clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME, &retry_time, NULL);
    posix_trace_get_status(trid,&status);
  }

  // OK. Flush (if any) is now over. Shut the stream down:
  posix_trace_shutdown(trid);

  return (void *) 0;
}




int init_module(void) {

  trace_attr_t   attr;
  int            error;
  int            i;
  trace_event_id_t event_string_id;
  trace_event_set_t  set1;
  char           fifo_name[16];
  event_data_struct event_data;

  // Open the fifos
  sprintf(fifo_name, "/dev/rtf%d", fifo1);
  rtf_create(fifo1, 100000);
  if ((fifo1_desc = open(fifo_name, O_NONBLOCK)) < 0) {
    rtl_printf("%s: [BROK]:Cannot open %s for writing\n", 
	       TEST_NAME, fifo_name);
    return -1;
  }
  sprintf(fifo_name, "/dev/rtf%d", fifo2);
  rtf_create(fifo2, 100000);
  if ((fifo2_desc = open(fifo_name, O_NONBLOCK)) < 0) {
    rtl_printf("%s: [BROK]:Cannot open %s for writing\n", 
	       TEST_NAME, fifo_name);
    return -1;
  }
  sprintf(fifo_name, "/dev/rtf%d", fifo3);
  rtf_create(fifo3, 100000);
  if ((fifo3_desc = open(fifo_name, O_NONBLOCK)) < 0) {
    rtl_printf("%s: [BROK]:Cannot open %s for writing\n", 
	       TEST_NAME, fifo_name);
    return -1;
  }
  sprintf(fifo_name, "/dev/rtf%d", fifo4);
  rtf_create(fifo4, 100000);
  if ((fifo4_desc = open(fifo_name, O_NONBLOCK)) < 0) {
    rtl_printf("%s: [BROK]:Cannot open %s for writing\n", 
	       TEST_NAME, fifo_name);
    return -1;
  }


  // Create the trace attribute:
  posix_trace_attr_init(&attr);

  // Set the attributes with values defined in common.h:
  posix_trace_attr_setstreamfullpolicy(&attr, MY_STREAM_POLICY);
  posix_trace_attr_setname(&attr, MY_STREAM_NAME);
  posix_trace_attr_setmaxdatasize(&attr, MY_EVENT_DATA_SIZE);
  posix_trace_attr_setstreamsize(&attr, MY_STREAM_SIZE);

  // Create the streams without log:
  if ((error = posix_trace_create(0, &attr, &trid1)) != 0) {
    rtl_printf("%s: [BROK]:posix_trace_create returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }
  if ((error = posix_trace_create(0, &attr, &trid2)) != 0) {
    rtl_printf("%s: [BROK]:posix_trace_create returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }
  if ((error = posix_trace_create(0, &attr, &trid3)) != 0) {
    rtl_printf("%s: [BROK]:posix_trace_create returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }
  if ((error = posix_trace_create(0, &attr, &trid4)) != 0) {
    rtl_printf("%s: [BROK]:posix_trace_create returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }

  // Create the streams with log:
  if ((error = posix_trace_create_withlog(0, &attr, 
					  fifo1_desc, &trid5)) != 0) {
    rtl_printf("%s: [BROK]:posix_trace_create_withlog returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }
  if ((error = posix_trace_create_withlog(0, &attr, 
					  fifo2_desc, &trid6)) != 0) {
    rtl_printf("%s: [BROK]:posix_trace_create_withlog returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }
  if ((error = posix_trace_create_withlog(0, &attr, 
					  fifo3_desc, &trid7)) != 0) {
    rtl_printf("%s: [BROK]:posix_trace_create_withlog returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }
  if ((error = posix_trace_create_withlog(0, &attr, 
					  fifo4_desc, &trid8)) != 0) {
    rtl_printf("%s: [BROK]:posix_trace_create_withlog returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }


  // Create a new user event type:
  if ((error = posix_trace_eventid_open (MY_USER_EVENT_NAME, 
					 &event_string_id)) != 0) {
    rtl_printf("%s: [BROK]:posix_trace_eventid_open returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }


  // Filter all *system* event types from the 8 streams:
  posix_trace_eventset_fill(&set1, POSIX_TRACE_SYSTEM_EVENTS);

  if ((error = posix_trace_set_filter(trid1, &set1, 
				      POSIX_TRACE_SET_EVENTSET)) != 0) { 
    rtl_printf("%s: [BROK]:posix_trace_get_filter returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }
  if ((error = posix_trace_set_filter(trid2, &set1, 
				      POSIX_TRACE_SET_EVENTSET)) != 0) { 
    rtl_printf("%s: [BROK]:posix_trace_get_filter returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }
  if ((error = posix_trace_set_filter(trid3, &set1, 
				      POSIX_TRACE_SET_EVENTSET)) != 0) { 
    rtl_printf("%s: [BROK]:posix_trace_get_filter returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }
  if ((error = posix_trace_set_filter(trid4, &set1, 
				      POSIX_TRACE_SET_EVENTSET)) != 0) { 
    rtl_printf("%s: [BROK]:posix_trace_get_filter returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }

  if ((error = posix_trace_set_filter(trid5, &set1, 
				      POSIX_TRACE_SET_EVENTSET)) != 0) { 
    rtl_printf("%s: [BROK]:posix_trace_get_filter returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }
  if ((error = posix_trace_set_filter(trid6, &set1, 
				      POSIX_TRACE_SET_EVENTSET)) != 0) { 
    rtl_printf("%s: [BROK]:posix_trace_get_filter returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }
  if ((error = posix_trace_set_filter(trid7, &set1, 
				      POSIX_TRACE_SET_EVENTSET)) != 0) { 
    rtl_printf("%s: [BROK]:posix_trace_get_filter returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }
  if ((error = posix_trace_set_filter(trid8, &set1, 
				      POSIX_TRACE_SET_EVENTSET)) != 0) { 
    rtl_printf("%s: [BROK]:posix_trace_get_filter returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }


  // Start the tracing of events: 
  if ((error = posix_trace_start(trid1)) != 0) {
    rtl_printf("%s: [BROK]:posix_trace_start returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }
  if ((error = posix_trace_start(trid2)) != 0) {
    rtl_printf("%s: [BROK]:posix_trace_start returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }
  if ((error = posix_trace_start(trid3)) != 0) {
    rtl_printf("%s: [BROK]:posix_trace_start returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }
  if ((error = posix_trace_start(trid4)) != 0) {
    rtl_printf("%s: [BROK]:posix_trace_start returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }
  if ((error = posix_trace_start(trid5)) != 0) {
    rtl_printf("%s: [BROK]:posix_trace_start returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }
  if ((error = posix_trace_start(trid6)) != 0) {
    rtl_printf("%s: [BROK]:posix_trace_start returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }
  if ((error = posix_trace_start(trid7)) != 0) {
    rtl_printf("%s: [BROK]:posix_trace_start returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }
  if ((error = posix_trace_start(trid8)) != 0) {
    rtl_printf("%s: [BROK]:posix_trace_start returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }


  // Trace an exact number of events:
  strcpy(event_data.str, MY_STRING);
  for (i = 0; i < NUMBER_OF_TRACED_EVENTS; i++) {
    event_data.serial = i;
    posix_trace_event(event_string_id, &event_data, sizeof(event_data));
  }


  // Stop the tracing of events: 
  if ((error = posix_trace_stop(trid1)) != 0) {
    rtl_printf("%s: [BROK]:posix_trace_stop returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }
  if ((error = posix_trace_stop(trid2)) != 0) {
    rtl_printf("%s: [BROK]:posix_trace_stop returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }
  if ((error = posix_trace_stop(trid3)) != 0) {
    rtl_printf("%s: [BROK]:posix_trace_stop returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }
  if ((error = posix_trace_stop(trid4)) != 0) {
    rtl_printf("%s: [BROK]:posix_trace_stop returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }
  if ((error = posix_trace_stop(trid5)) != 0) {
    rtl_printf("%s: [BROK]:posix_trace_stop returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }
  if ((error = posix_trace_stop(trid6)) != 0) {
    rtl_printf("%s: [BROK]:posix_trace_stop returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }
  if ((error = posix_trace_stop(trid7)) != 0) {
    rtl_printf("%s: [BROK]:posix_trace_stop returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }
  if ((error = posix_trace_stop(trid8)) != 0) {
    rtl_printf("%s: [BROK]:posix_trace_stop returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }


  // Create the 4 tasks to retrieve on-line from the streams w/o log:
  pthread_create (&reader_task1, NULL, task_reader, (void *) &trid1);
  pthread_create (&reader_task2, NULL, task_reader, (void *) &trid2);
  pthread_create (&reader_task3, NULL, task_reader, (void *) &trid3);
  pthread_create (&reader_task4, NULL, task_reader, (void *) &trid4);


  // Create the 4 tasks to shutdown the 4 streams with log:
  pthread_create (&shutdown_task1, NULL, task_to_shutdown, (void *) &trid5); 
  pthread_create (&shutdown_task2, NULL, task_to_shutdown, (void *) &trid6); 
  pthread_create (&shutdown_task3, NULL, task_to_shutdown, (void *) &trid7); 
  pthread_create (&shutdown_task4, NULL, task_to_shutdown, (void *) &trid8); 

  if (! test_failed) {
    rtl_printf("%s: PASS :Online/offline trace stress passed. Logs generated.\n", 
	       TEST_NAME);  
  }
  return test_failed;

}





void cleanup_module(void) {

  // Cancel the shutdown tasks
  pthread_cancel(shutdown_task1);
  pthread_cancel(shutdown_task2);
  pthread_cancel(shutdown_task3);
  pthread_cancel(shutdown_task4);

  pthread_join(shutdown_task1, NULL);
  pthread_join(shutdown_task2, NULL);
  pthread_join(shutdown_task3, NULL);
  pthread_join(shutdown_task4, NULL);

  // Close the fifos
  close(fifo1_desc);
  close(fifo2_desc);
  close(fifo3_desc);
  close(fifo4_desc);

  rtf_destroy(fifo1);
  rtf_destroy(fifo2);
  rtf_destroy(fifo3);
  rtf_destroy(fifo4);

}
