
/*

   FILE:

	wlog.c
								
   DESCRIPTION: 
 
         This simple program opens a device file and, from it,
         retrieves raw (POSIX) trace events from the RT-Linux
         scheduler and writes them into a log. This log can be opened
         afterwards with the regular POSIX trace interface into a
         pre-recorded stream.

   Copyright (C) 1998
	Ana Garcia Fornes
	Andres Terrasa Barrena
	Ignacio Paches Giner
	Agustin Espinosa Minguet



  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software 
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.


  LAST REVISION:
        December 2003
*/


#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include <trace.h>
#include <posix/trace/comm.h>


static void _do_copy (int input, int output, int nbytes) {
  char           data_chunk[1024];
  int            count;
  int            ret;

  count = (nbytes > 1024 ? 1024 : nbytes);

  while (nbytes > 0) {

    if ((ret = read(input, &data_chunk, count)) < 0) {
      fprintf(stderr, "wlog: bad data while reading input file\n");
      exit(1);
    }
    
    if ((ret = write(output, &data_chunk, count)) < 0) {
      fprintf(stderr, "wlog: unable to write to log file\n");
      exit(1);
    }

    nbytes -= count;
  }
}




int main (int argc, char *argv[]) {

  char           input_file[256];
  char           log_file[256];
  int            input_fd;
  int            log_fd;
  flush_header_t flush_info;
  int            ret;
  off_t          pos;
  trace_attr_t   attr;
  size_t         header_size;
  int            header_read_ok;
  int            done;
  posix_trace_status_info  status;
  event_type_info_t  event_types[TRACE_EVENT_MAX];

  // Command-line arguments control:
  if (argc < 3) {
    fprintf(stderr, "Use: wlog inputfile logfile\n");
    exit(1);
  } else if (argc == 3) {
    strncpy(input_file, argv[1], sizeof(input_file));
    strncpy(log_file, argv[2], sizeof(log_file));
  }

  // Open input (device)  file and (output) log:
  if ((input_fd = open(input_file, O_RDONLY | O_SYNC)) == -1) {
    fprintf(stderr, "wlog: Cannot open %s for reading\n", input_file);
    exit(1);
  }

  if ((log_fd = open(log_file, O_RDWR | O_CREAT | O_TRUNC,
		     S_IRUSR | S_IWUSR |  S_IRGRP | S_IROTH)) == -1) {
    fprintf(stderr, "wlog: Cannot create/open %s for writing\n", log_file);
    exit(1);
  }


  // Read the first data from the input file:
  ret = read(input_fd, &flush_info, sizeof(flush_info));
  if (ret < 0 || flush_info.type != FLUSH_LOG_STARTING) {
    fprintf(stderr, "wlog: bad data while reading input file\n");
    exit(1);
  }


  /*
    Now prepare the log file header: write the initial magic and leave
    an space enough for writing the log header: attributes, event
    identifiers and final status.
  */
  write(log_fd, &(flush_info.type), sizeof(int));

  header_size = sizeof(int) + 
    sizeof(trace_attr_t) + 
    (sizeof(event_type_info_t) * TRACE_EVENT_MAX) +
    sizeof(posix_trace_status_info);
  
  pos = lseek(log_fd, (off_t) header_size, SEEK_SET);

  if (pos < 0) {
    fprintf(stderr, "wlog: unable to write header to log file\n");
    exit(1);
  }


  /* 
     Main loop : read input and write log until work is done (or an
     interrupt kills us)
  */
  header_read_ok = 0;
  done = 0;
  while (! done) {

    ret = read(input_fd, &flush_info, sizeof(flush_info));

#ifdef _WLOG_DEBUG
    fprintf(stderr, "Message read. Type = %x\n", flush_info.type);
#endif

    switch (flush_info.type) {

    case FLUSH_LOG_EVENT : 
      _do_copy(input_fd, log_fd, flush_info.add_bytes);
      break;

    case FLUSH_LOG_HEADER :
      ret = read(input_fd, &attr, sizeof(trace_attr_t));
      if (ret < sizeof(trace_attr_t)) {
	fprintf(stderr, "wlog: incomplete attr while reading input file\n");
	exit(1);
      }

      ret = read(input_fd, &event_types, 
		 sizeof(event_type_info_t) * TRACE_EVENT_MAX);
      if (ret < (sizeof(event_type_info_t) * TRACE_EVENT_MAX)) {
	fprintf(stderr, "wlog: incomplete ev types while reading input file\n");
	fprintf(stderr, "wlog: Read %d bytes while expecting %d\n",
		(int) ret, (int) (sizeof(event_type_info_t) * TRACE_EVENT_MAX));
	exit(1);
      }

      ret = read(input_fd, &status, sizeof(posix_trace_status_info));
      if (ret < sizeof(posix_trace_status_info)) {
	fprintf(stderr, "wlog: incomplete status  while reading input file\n");
	exit(1);
      }

      header_read_ok = 1;
      break;

    case FLUSH_LOG_DONE :
      if (! header_read_ok) {
	fprintf(stderr, "wlog: LOG_DONE before complete header\n");
	exit(1);
      }

      // Go to the header section and write the header's members:
      pos = lseek(log_fd, (off_t) sizeof(int), SEEK_SET);
      if (pos < 0) {
	fprintf(stderr, "wlog: unable to lseek to log file's header\n");
	exit(1);
      }

      ret = write(log_fd, &attr, sizeof(trace_attr_t));
      if (ret < 0) {
	fprintf(stderr, "wlog: unable to write attr to log file's header\n");
	exit(1);
      }

      ret = write(log_fd, &event_types, 
		  sizeof(event_type_info_t) * TRACE_EVENT_MAX);
      if (ret < 0) {
	fprintf(stderr, "wlog: unable to write event types to log file's header\n");
	exit(1);
      }

      ret = write(log_fd, &status, sizeof(posix_trace_status_info));
      if (ret < 0) {
	fprintf(stderr, "wlog: unable to write status to log file's header\n");
	exit(1);
      }
      
      // Finally, go to the beginning of the file and write the "ok" magic:
      pos = lseek(log_fd, (off_t) 0, SEEK_SET);
      if (pos < 0) {
	fprintf(stderr, "wlog: unable to lseek to log file's end\n");
	exit(1);
      }

      ret = write(log_fd, &(flush_info.type), sizeof(int));
      if (ret < 0) {
	fprintf(stderr, "wlog: unable to write magic to log file's end\n");
	exit(1);
      }

      // We're all set, so:
      done = 1;
      break;

    default:
      fprintf(stderr, "wlog: unknown header while reading input file\n");
      exit(1);
    }
  }


  // All done. Close everything up and go...
  close(input_fd);
  fsync(log_fd);
  close(log_fd);
  exit(0);
  return 0;
}
